"""
Window class.
"""

# =============================================================================

import math
import tkinter as _tk
import tkinter.font as _TkFont
from pathlib import Path
from typing import Dict, Optional, Tuple

try:
    from PIL import Image, ImageTk
except ImportError as ex:
    raise RuntimeError(
        """\
Pillow must be installed to run a game in a window:

  $ python -m pip install Pillow
"""
    ) from ex

from game_state import GameState
from pieces import Piece
from position import BoardPosition, Position
from utils import Color, PieceType, PromoteType, check_brc_bool

# =============================================================================

# The size of a square cell.
SQUARE_SIZE = 70
HALF_SQUARE_SIZE = SQUARE_SIZE / 2
# The padding between the two boards.
BOARD_PADDING = 50
# The offset from the edges of the board to a label.
LABEL_OFFSET = HALF_SQUARE_SIZE
# The offset to the top-left corner of Board A.
X_OFFSET = 70
Y_OFFSET = 70
# The offset to the left side of Board B.
X_OFFSET2 = X_OFFSET + SQUARE_SIZE * 8 + BOARD_PADDING
# The offset from the bottom of the board to the bottom of the screen.
Y_OFFSET_BOT = 180

# Calculated dimensions of the entire window.
WIDTH = X_OFFSET + SQUARE_SIZE * 8 + BOARD_PADDING + SQUARE_SIZE * 8 + X_OFFSET
HALF_WIDTH = WIDTH / 2
HEIGHT = Y_OFFSET + SQUARE_SIZE * 8 + Y_OFFSET_BOT

RESET_BUTTON_COORDS = (10, 10, 70, 40)
UNDO_BUTTON_COORDS = (WIDTH - 70, 10, WIDTH - 10, 40)
POSSIBLE_DOT_RADIUS = 5

# Board colors
WHITE_COLOR = "white"
BLACK_COLOR = "gray"
SELECT_COLOR = "green"
POSSIBLE_COLOR = "blue"
LAST_MOVE_COLOR = "#" + "".join(f"{x:02x}" for x in (102, 255, 178))

# The fonts to use.
FONTS = ("SF Pro",)

# Images
PICTURE_FOLDER = Path("pictures")
PICTURE_EXT = ".png"

# =============================================================================


def _black_on_button(game_state: GameState) -> bool:
    # black should only be on the bottom if white is not a human but
    # black is a human
    # otherwise, default to white on bottom
    return not game_state.white.is_human and game_state.black.is_human


def _brc_to_xy(
    game_state: GameState, pos: BoardPosition, *, center: bool
) -> Tuple[int, int]:
    """Calculates the xy-coordinates of the top-left or center of the
    square for the given board position.
    """
    bn, r, c = pos
    if _black_on_button(game_state):
        # flip the coordinates
        r = 7 - r
        c = 7 - c
    x = (X_OFFSET, X_OFFSET2)[bn] + SQUARE_SIZE * c
    y = Y_OFFSET + SQUARE_SIZE * r
    if center:
        x += HALF_SQUARE_SIZE
        y += HALF_SQUARE_SIZE
    return x, y


def _xy_to_brc(
    game_state: GameState, x: int, y: int
) -> Optional[BoardPosition]:
    """Calculates the board position of the given xy-coordinates."""
    bn = 0
    # int truncates (toward 0), so must use `math.floor()` to account
    # for possible negative numbers
    r = math.floor((y - Y_OFFSET) / SQUARE_SIZE)
    c = math.floor((x - X_OFFSET) / SQUARE_SIZE)
    if c >= 8:  # try other board
        bn = 1
        c = math.floor((x - X_OFFSET2) / SQUARE_SIZE)
    if _black_on_button(game_state):
        # flip the coordinates
        r = 7 - r
        c = 7 - c
    if not check_brc_bool(bn, r, c):
        # invalid coords (not within boards)
        return None
    return BoardPosition(bn, r, c)


def _within_coords(coords, x, y):
    x1, y1, x2, y2 = coords
    return x1 <= x <= x2 and y1 <= y <= y2


# =============================================================================


class Square:
    """Represents a square cell on the board.

    Methods:
        reset(): Resets this square.
        select(): Selects this square.
        last_move(): Highlights this square as the last move.
        possible(): Sets this square as a possible move.
    """

    def __init__(
        self,
        game_state: GameState,
        canvas: _tk.Canvas,
        bn: int,
        r: int,
        c: int,
    ):
        self._canvas = canvas
        self._color = WHITE_COLOR if (r + c) % 2 == 0 else BLACK_COLOR

        x, y = _brc_to_xy(game_state, (bn, r, c), center=False)
        self._square = canvas.create_rectangle(
            x, y, x + SQUARE_SIZE, y + SQUARE_SIZE, fill=self._color
        )
        x_mid = x + HALF_SQUARE_SIZE
        y_mid = y + HALF_SQUARE_SIZE
        self._dot = canvas.create_oval(
            x_mid - POSSIBLE_DOT_RADIUS,
            y_mid - POSSIBLE_DOT_RADIUS,
            x_mid + POSSIBLE_DOT_RADIUS,
            y_mid + POSSIBLE_DOT_RADIUS,
            fill=POSSIBLE_COLOR,
            tags=("possible_dot",),
            state="hidden",
        )

    def _set_color(self, color: str):
        self._canvas.itemconfig(self._square, fill=color)

    def reset(self):
        """Resets this square."""
        self._set_color(self._color)
        self.reset_possible()

    def select(self):
        """Selects this square."""
        self._set_color(SELECT_COLOR)

    def last_move(self):
        """Highlights this square as the last move."""
        self._set_color(LAST_MOVE_COLOR)

    def possible(self):
        """Sets this square as a possible move."""
        self._canvas.itemconfig(self._dot, state="normal")

    def reset_possible(self):
        """Hides this square's possible move dot."""
        self._canvas.itemconfig(self._dot, state="hidden")


class Window:
    """A window where the game is played.

    Methods:
        run(): Runs the game.
    """

    def __init__(
        self,
        game_state: GameState,
        title: str = "Alice Chess",
        non_human_player_delay: int = 3,
    ):
        """Initializes a Window.

        Args:
            game_state (GameState): The starting game state.
            title (str): The title of the window.
            non_human_player_delay (int): The number of seconds of delay
                for non-human players to simulate realism.

        Raises:
            ValueError:
                If `non_human_player_delay` is negative.
            FileNotFoundError:
                If the pictures folder does not exist.
                If a piece image file is not found.
        """
        if not PICTURE_FOLDER.exists():
            raise FileNotFoundError(
                f"Pictures folder does not exist: {PICTURE_FOLDER}"
            )

        if non_human_player_delay < 0:
            raise ValueError("`non_human_player_delay` cannot be negative")
        self._delay = non_human_player_delay * 1000

        self._game = game_state
        white = game_state.white
        black = game_state.black
        self._has_human_player = white.is_human or black.is_human
        self._undo_state = None
        self._find_undo_state()

        if _black_on_button(game_state):

            def label(i):
                num = str(8 - i)
                letter = "abcdefgh"[7 - i]
                return num, letter

        else:

            def label(i):
                num = str(i + 1)
                letter = "abcdefgh"[i]
                return num, letter

        self._tk = _tk.Tk()
        self._tk.title(title)

        # find font to use
        self._font = None
        families = set(_TkFont.families())
        for font in FONTS:
            if font in families:
                self._font = font
                break
        font30 = (self._font, 30)
        font20 = (self._font, 20)
        font15 = (self._font, 15)

        self._canvas = canvas = _tk.Canvas(
            self._tk, width=WIDTH, height=HEIGHT
        )
        canvas.pack()

        # title
        canvas.create_text(HALF_WIDTH, Y_OFFSET / 2, text=title, font=font30)

        # piece images
        # 5:4 scale for image:square sizes
        img_size = (int(SQUARE_SIZE * 5 / 4),) * 2
        # maps: color -> piece type -> image object
        self._images: Dict[Color, Dict[PieceType, ImageTk.PhotoImage]] = {}
        for color in Color:
            color_images = {}
            for piece_type in PieceType:
                image_path = (
                    PICTURE_FOLDER
                    / f"{color.abbr()}{piece_type.value}{PICTURE_EXT}"
                )
                if not image_path.exists():
                    raise FileNotFoundError(
                        f"Piece image file not found: {image_path}"
                    )
                img = Image.open(image_path)
                img = img.resize(img_size, Image.ANTIALIAS)
                color_images[piece_type] = ImageTk.PhotoImage(img)
            self._images[color] = color_images

        # maps: piece id -> image id
        self._piece_images: Dict[int, int] = {}

        # draw boards
        self._squares: Dict[BoardPosition, Square] = {}
        for bn in range(2):
            for r in range(8):
                for c in range(8):
                    self._squares[bn, r, c] = Square(
                        game_state, canvas, bn, r, c
                    )

        # labels
        board1_x = X_OFFSET - LABEL_OFFSET
        board2_x = X_OFFSET2 + LABEL_OFFSET
        board2_end_x = WIDTH - X_OFFSET + LABEL_OFFSET
        under_y = HEIGHT - Y_OFFSET_BOT + LABEL_OFFSET
        for i in range(8):
            num, letter = label(i)
            # row labels (numbers)
            y = under_y - SQUARE_SIZE * (i + 1)
            canvas.create_text(board1_x, y, text=num, font=font20)
            canvas.create_text(board2_end_x, y, text=num, font=font20)
            # column labels (letters)
            x1 = board1_x + SQUARE_SIZE * (i + 1)
            x2 = board2_x + SQUARE_SIZE * i
            canvas.create_text(x1, under_y, text=letter, font=font20)
            canvas.create_text(x2, under_y, text=letter, font=font20)

        # between the center of the label text and the bottom of the
        # window
        bottom_y = HEIGHT - (Y_OFFSET_BOT - LABEL_OFFSET) / 2

        # promotions (in same order as `PromoteType`)
        # center the 4 pieces around the middle
        pro_x = HALF_WIDTH - SQUARE_SIZE - HALF_SQUARE_SIZE
        for color in Color:
            tag = f"{color.name.lower()}_promotions"
            x = pro_x
            for promote_type in PromoteType:
                piece_type = promote_type.value
                canvas.create_image(
                    x,
                    bottom_y,
                    image=self._images[color][piece_type],
                    tags=(tag, "promotions"),
                )
                x += SQUARE_SIZE
        self._hide("promotions")

        # state text
        self._state_text = canvas.create_text(
            HALF_WIDTH, bottom_y, font=font20
        )

        # reset button
        canvas.create_rectangle(*RESET_BUTTON_COORDS)
        x1, y1, x2, y2 = RESET_BUTTON_COORDS
        canvas.create_text(
            (x1 + x2) / 2, (y1 + y2) / 2, text="Reset", font=font15
        )

        # undo button
        canvas.create_rectangle(*UNDO_BUTTON_COORDS, tags=("undo",))
        x1, y1, x2, y2 = UNDO_BUTTON_COORDS
        canvas.create_text(
            (x1 + x2) / 2,
            (y1 + y2) / 2,
            text="Undo",
            font=font15,
            tags=("undo",),
        )
        self._hide("undo")

        # for manually playing with human players
        self._last_move = None
        self._selected = None
        self._other_player_callback = None

    def _show(self, element, **kwargs):
        self._canvas.itemconfig(element, **kwargs, state="normal")

    def _hide(self, element, **kwargs):
        self._canvas.itemconfig(element, **kwargs, state="hidden")

    def _stop_non_human_player(self):
        """Stops any non-human player from thinking."""
        # stop any non-human player from thinking
        if self._other_player_callback is not None:
            self._canvas.after_cancel(self._other_player_callback)
            self._other_player_callback = None

    def _reset(self):
        """Resets the window."""

        # stop any non-human player from thinking
        self._stop_non_human_player()

        # restart the game
        self._game = self._game.restart()

        # reset squares
        self._last_move = None
        self._selected = None
        for square in self._squares.values():
            square.reset()

        # reset other visual stuff
        self._hide("undo")
        self._hide("promotions")
        self._hide(self._state_text)

        # update piece images
        self._update_pieces()

    def _update_piece_image(
        self,
        piece: Piece,
        img_id: Optional[int] = None,
        pos: Optional[Position] = None,
    ):
        if pos is None:
            pos = piece.pos
        x, y = _brc_to_xy(self._game, pos, center=True)
        if img_id is None:
            # create new image
            img_id = self._canvas.create_image(
                x, y, image=self._images[piece.color][piece.type]
            )
        else:
            # move image to proper location
            self._canvas.coords(img_id, x, y)
            # update the actual image to be the proper piece type
            self._canvas.itemconfig(
                img_id, image=self._images[piece.color][piece.type]
            )
        return img_id

    def _update_pieces(self):
        """Updates the piece images."""
        piece_images = {}
        for piece in self._game.yield_all_pieces():
            img_id = self._piece_images.pop(piece.id, None)
            piece_images[piece.id] = self._update_piece_image(piece, img_id)
        for img_id in self._piece_images.values():
            self._canvas.delete(img_id)
        self._piece_images = piece_images

        # make sure all the "possible move" dots are above the images
        self._canvas.lift("possible_dot")

    def _unselect_piece(self):
        if self._selected is None:
            return
        self._squares[self._selected.pos].reset()
        for move in self._selected.yield_moves():
            self._squares[move.capture_pos].reset_possible()
        self._selected = None

    def _select_piece(self, piece: Piece):
        print("selected piece:", piece)
        self._unselect_piece()
        self._selected = piece
        self._squares[self._selected.pos].select()
        for move in self._selected.yield_moves():
            self._squares[move.capture_pos].possible()

    def _find_undo_state(self):
        # go back until find a state where it's a human player's turn
        self._undo_state = None
        if not self._has_human_player:
            return
        state = self._game
        prev = state.prev
        if prev is None:
            # we're at the first state in the game; can't undo
            pass
        elif prev.current_player.is_human:
            # undo to this state
            self._undo_state = prev
        else:
            # undo to the previous state
            grand_prev = prev.prev
            if grand_prev is None:
                # we're at the second state; can't undo
                pass
            else:
                self._undo_state = grand_prev

    def _click(self, event: _tk.Event):
        """Handles a click event."""
        x, y = event.x, event.y

        # check reset button
        if _within_coords(RESET_BUTTON_COORDS, x, y):
            self._reset()
            # initialize a non-human turn, if possible
            self._non_human_turn(True)
            return

        if not self._has_human_player:
            # no other clicks to check for
            return

        # check undo button
        if _within_coords(UNDO_BUTTON_COORDS, x, y):
            if self._undo_state is not None:
                self._stop_non_human_player()
                self._game = self._undo_state
                self._update_pieces()
                self._end_turn()
            return

        # check game over
        if self._game.is_game_over():
            return

        if self._game.needs_promotion():
            pro_x = HALF_WIDTH - SQUARE_SIZE * 2
            pro_y = HEIGHT - Y_OFFSET_BOT / 2 - HALF_SQUARE_SIZE
            if not pro_y <= y <= pro_y + SQUARE_SIZE:
                return
            i = int((x - pro_x) / SQUARE_SIZE)
            if not 0 <= i < 4:
                return
            promote_type = PromoteType.by_index(i)
            print("promoting pawn to:", promote_type)
            self._game = self._game.promote(promote_type)
            self._hide("promotions")
            self._end_turn()
            return

        # if not a human player's turn, ignore
        if not self._game.current_player.is_human:
            return

        clicked_pos = _xy_to_brc(self._game, x, y)
        if clicked_pos is None:
            # invalid coords
            return

        piece = self._game.get_piece(*clicked_pos)

        # selecting a new piece
        if self._selected is None:
            if piece is None:
                return
            if piece.color is not self._game.current_color:
                return
            print("=" * 15)
            print("clicked:", clicked_pos, clicked_pos.pos)
            self._select_piece(piece)
            return

        # selecting a different piece
        if piece is not None and piece.color is self._game.current_color:
            if piece is self._selected:
                # clicked same piece; unselect it
                self._unselect_piece()
            else:
                print("=" * 15)
                print("clicked:", clicked_pos, clicked_pos.pos)
                self._select_piece(piece)
            return

        if not self._selected.can_make_move(*clicked_pos):
            # invalid move for selected piece
            return

        # make move
        print("=" * 15)
        print("moving selected to:", clicked_pos, clicked_pos.pos)
        try:
            _, r, c = clicked_pos
            move = (self._selected.pos, (r, c))
            self._game = self._game.make_move(move)
        except ValueError as e:
            print(" ", "ERROR:", e)
            return

        # if needs promotion, wait
        if self._game.needs_promotion():
            pawn = self._selected
            # move the pawn forward to the clicked position
            self._piece_images[pawn.id] = self._update_piece_image(
                pawn, self._piece_images.get(pawn.id, None), pos=clicked_pos
            )
            # delete the captured piece
            if self._game.move.move_captured:
                img_id = self._piece_images.pop(
                    self._game.move.piece_captured.id, None
                )
                if img_id is not None:
                    self._canvas.delete(img_id)
            # reset the piece's possible dots
            for move in pawn.yield_moves():
                self._squares[move.capture_pos].reset_possible()
            # show promotion images
            self._show(f"{self._selected.color.name.lower()}_promotions")
            return

        self._end_turn()

    def _non_human_turn(self, first_move: bool = False):
        if self._game.current_player.is_human:
            return

        def make_turn():
            self._game = self._game.step()
            print("non-human player made move:", self._game.move)
            self._other_player_callback = None
            self._end_turn()

        if not first_move and self._delay > 0:
            if self._game.is_in_check():
                action = "is in check. Thinking..."
            else:
                action = "thinking..."
            text = f"{self._game.current_color.title()} {action}"
            self._show(self._state_text, text=text)
            self._other_player_callback = self._canvas.after(
                self._delay, make_turn
            )
        else:
            make_turn()

    def _end_turn(self):
        self._hide(self._state_text, text="")

        self._find_undo_state()
        if self._undo_state is not None:
            self._show("undo")
        else:
            self._hide("undo")

        self._update_pieces()

        self._unselect_piece()
        if self._last_move is not None:
            self._squares[self._last_move].reset()

        # highlight the last move
        if self._game.prev is None:
            # no last move
            self._last_move = None
        else:
            last_pos = self._game.move.result_pos
            self._squares[last_pos].last_move()
            self._last_move = last_pos

        game = self._game
        game.debug()
        # check for game over
        if game.is_game_over():
            # no more undo
            # self._hide("undo")
            if game.is_in_checkmate():
                text = f"Checkmate. {game.winner().title()} won!"
            elif game.is_in_stalemate():
                text = "Stalemate."
            elif game.is_kings_draw():
                text = "Draw."
            elif game.is_draw():
                text = "Draw (50 moves rule)."
            self._show(self._state_text, text=text)
            return

        # check for check
        if game.is_in_check():
            self._show(
                self._state_text,
                text=f"{game.current_color.title()} is in check.",
            )

        # initialize next turn
        self._non_human_turn()

    def run(self):
        """Runs the game."""
        self._reset()

        # bind clicks
        self._canvas.bind_all("<Button-1>", self._click)

        # initialize a non-human turn, if possible
        # if no human players, this will keep going forever until the
        # game is over
        # use an `after` call so that the mainloop is still entered even
        # if there are no human players
        self._canvas.after(0, self._non_human_turn, True)

        self._tk.mainloop()



"""
Piece abstract class.
"""

# =============================================================================

from typing import Iterator, Optional, Set

from position import BoardPosition, PieceMove, Position
from utils import Color, PieceType, check_brc

# =============================================================================

__all__ = ("Piece",)

# =============================================================================


class Piece:
    """Represents a piece.

    Properties:
        id (int): The piece id.
        name (str): The piece name.
        fen_name (str): The piece FEN name.
        type (PieceType): The piece type.
        color (Color): The piece color.
        has_moved (bool): Whether this piece has moved.
        pos (BoardPosition): The piece position.
        num_moves (int): The number of possible moves.
        is_captured (bool): Whether this piece is captured.
        is_threatened (bool): Whether this piece is being threatened on
            the current board. False if not known.

    Methods:
        copy() -> Piece:
            Returns a copy of this piece.

        yield_moves() -> Iterator[PieceMove]
            Yields the possible moves for this piece.
        can_make_move(bn, tr, tc) -> bool
            Returns whether this piece can move to the given position.

        move_to(pos) -> Piece
            Returns a copy of this piece that is moved to the given
            position.
        capture() -> Piece
            Returns a copy of this piece that is captured.
    """

    _type: PieceType = None

    def __init__(
        self,
        piece_id: int,
        color: Color,
        pos: Optional[BoardPosition] = None,
        captured: bool = False,
        has_moved: bool = False,
    ):
        if self._type is None:
            raise RuntimeError(
                f"{self.__class__.__name__} is missing `_type` attribute"
            )
        if pos is None and not captured:
            raise ValueError("Position not given for non-captured piece")
        if captured:
            pos = None

        self._id = piece_id
        self._color = color
        self._name = color.abbr() + self._type.value
        self._fen_name = self._type.value
        if self._color is Color.BLACK:
            self._fen_name = self._fen_name.lower()

        self._pos = BoardPosition.of(pos)
        self._has_moved = has_moved
        self._is_captured = captured
        self._is_threatened = False

        self._moves = None

    def __str__(self):
        attrs = {
            "id": self._id,
            "color": self._color.name,
            "pos": self._pos,
        }
        if self._is_captured:
            attrs["captured"] = True
        attrs_str = " ".join(f"{key}={val}" for key, val in attrs.items())
        return f"{self._type.title()}<{attrs_str}>"

    def __repr__(self):
        args = {
            "piece_id": self._id,
            "color": self._color,
            "pos": repr(self._pos),
            "captured": self._is_captured,
        }
        args_str = ", ".join(f"{key}={val}" for key, val in args.items())
        return f"{self.__class__.__name__}({args_str})"

    @property
    def id(self) -> int:
        """The piece id."""
        return self._id

    @property
    def name(self) -> str:
        """The piece name."""
        return self._name

    @property
    def fen_name(self) -> str:
        """The piece FEN name."""
        return self._fen_name

    @property
    def type(self) -> PieceType:
        """The piece type."""
        return self._type

    @property
    def color(self) -> Color:
        """The piece color."""
        return self._color

    def _check_start_position(self) -> bool:
        """Checks if the current piece position could be a start
        position of the piece.
        """
        raise NotImplementedError()

    @property
    def has_moved(self) -> bool:
        """Whether this piece has moved."""
        if self._has_moved:
            return True
        # also check if this piece is in one of its starting positions
        return not self._check_start_position()

    @property
    def pos(self) -> BoardPosition:
        """The piece position."""
        return self._pos

    @property
    def num_moves(self) -> int:
        """The number of possible moves."""
        return len(self._get_moves())

    @property
    def is_captured(self) -> bool:
        """Whether this piece is captured."""
        return self._is_captured

    @property
    def is_threatened(self) -> bool:
        """Whether this piece is being threatened on the current board.

        False if not known.
        """
        return self._is_threatened

    def copy(self) -> "Piece":
        """Returns a copy of this piece."""
        return self.__class__(
            self._id,
            self._color,
            self._pos,
            self._is_captured,
            self._has_moved,
        )

    def _set_possible_moves(self, moves: Set[Position]):
        """Sets the possible moves of this piece. Should only be called
        by a `MovesCalculator` instance.
        """
        if self._is_captured:
            raise RuntimeError("Cannot set possible moves for captured piece")
        self._moves = []
        for target in moves:
            self._moves.append(PieceMove(self._pos, target, self))
        self._moves.sort()

    def _add_move(self, tr: int, tc: int):
        """Adds the given move."""
        if self._moves is None:
            raise ValueError(
                f"{self._type.title()} has not calculated its possible moves"
            )
        check_brc(tr=tr, tc=tc)
        self._moves.append(PieceMove(self._pos, (tr, tc), self))
        self._moves.sort()

    def _mark_threatened(self, is_threatened: bool):
        """Marks the current piece as being threatened or not. Should
        only be called by a `MovesCalculator` instance.
        """
        if self._is_captured:
            raise RuntimeError("Cannot mark threatened for a captured piece")
        self._is_threatened = is_threatened

    def _set_moved(self):
        """Sets this piece as having moved."""
        self._has_moved = True

    def _get_moves(self):
        if self._is_captured:
            raise RuntimeError("Cannot get moves for captured piece")
        if self._moves is None:
            raise RuntimeError(
                f"{self._type.title()} has not calculated its possible moves"
            )
        return self._moves

    def yield_moves(self) -> Iterator[PieceMove]:
        """Yields the possible moves for this piece."""
        return iter(self._get_moves())

    def can_make_move(self, bn: int, tr: int, tc: int) -> bool:
        """Returns whether this piece can move to the given position.

        Args:
            bn (int): The board number.
            tr (int): The target row.
            tc (int): The target column.

        Raises:
            ValueError:
                If `bn` is not 0 or 1.
                If `tr` and `tc` are not bounded within [0, 7].

        Returns:
            bool: Whether this piece can move to the given position.
        """
        check_brc(bn=bn, tr=tr, tc=tc)
        if bn != self._pos.bn:
            # can't move to somewhere on the other board
            return False
        return (tr, tc) in (move.target for move in self._get_moves())

    def move_to(self, pos: BoardPosition) -> "Piece":
        """Returns a copy of this piece that is moved to the given
        position.

        Also changes `has_moved` to True.
        """
        if self._is_captured:
            raise RuntimeError("Cannot move a captured piece")
        # can't be captured, so leave it as false
        return self.__class__(self._id, self._color, pos, has_moved=True)

    def capture(self) -> "Piece":
        """Returns a copy of this piece that is captured."""
        if self._is_captured:
            return self
        return self.__class__(
            self._id, self._color, captured=True, has_moved=self._has_moved
        )



from game_state import GameState
from player import HumanPlayer

# incorrect checkmate - fixed
fen = "rn3b1r/ppp1pp1p/8/6Q1/8/8/PPP1PPPP/R1B1KBNR/8/8/8/8/k2P4/2N5/8/3q4 b - - 8 39"

# testing
# fen = "k7/1N6/8/8/8/8/8/7K/8/8/8/8/8/8/8/8 b - - 0 0"
# fen = "k7/7R/8/8/8/8/8/1R5K/8/8/8/8/8/8/8/8 b - - 0 0"

GameState.from_fen(fen, white=HumanPlayer, black=HumanPlayer).debug()



"""
GameState class.
"""

# =============================================================================

import math
from itertools import count, zip_longest
from typing import Dict, Iterable, Iterator, Optional, Self, Type

import pieces
from moves_calculator import BoardDict, MovesCalculator
from pieces import Piece, make_promoted
from player import AnyPlayer
from position import Move, PieceMove, PieceMoved, Position
from utils import Color, EndGameState, PieceType, PromoteType

# =============================================================================

__all__ = ("GameState",)

# =============================================================================

_PRIVATE_game_state_constructor_key = object()


class GameState:  # pylint: disable=too-many-public-methods
    """Immutable state of a game.

    Includes board state, previous moves, captured pieces, possible
    moves, and end game state.

    Constructors:
        GameState.new(*, white, black)
            Initializes the first GameState of a game.
        GameState.from_fen(fen, *, white, black)
            Initializes a GameState from a FEN string, using the given
            players.

    Properties:
        id (int): The game state id.

        white (Player): The white player.
        black (Player): The black player.

        prev (Optional[GameState]): The previous GameState, or None if
            this is the first state.
        move (Optional[PieceMoved]): The move from the previous
            GameState, or None if this is the first state.

        current_color (Color): The color of the current player.
        current_player (Player): The current player.

        half_move_clock (int): The number of half-moves since the last
            capture or the last pawn moved.
        num_moves (int): The number of full moves played in the entire
            game.

    Methods:
        fen() -> str
            Returns the FEN representation of the game.
        board_to_str(empty=" .") -> str
            Returns a string representation of the board.
        moves_to_str(columns=4, captured_empty=True) -> str
            Returns a string representation of the moves for each piece.

        is_game_over() -> bool
            Returns whether the game is over (checkmate, stalemate, or
            draw).
        is_in_checkmate() -> bool
            Returns whether the current player is in checkmate.
        winner() -> Optional[Color]
            Returns the winner of the game (for checkmate).
        is_in_stalemate() -> bool
            Returns whether the current player is in stalemate.
        is_kings_draw() -> bool
            Returns whether the game is a draw by only having kings.
        is_draw() -> bool
            Returns whether the game is a draw (by the fifty move rule).

        is_in_check() -> bool
            Returns whether the current player is in check.
        needs_promotion() -> bool
            Returns whether the state is waiting for a promotion for the
            last moved pawn.
        promote(promote_type) -> GameState
            Promotes the promoting pawn.

        yield_all_pieces() -> Iterator[Piece]
            Yields all the pieces.
        yield_player_pieces() -> Iterator[Piece]
            Yields all the pieces for the current player.
        yield_player_moves() -> Iterator[PieceMove]
            Yields all the possible moves for the current player.
        num_captured() -> int
            Returns the number of captured pieces.
        yield_captured() -> Iterator[Piece]
            Yields the captured pieces.

        get_piece(bn, r, c) -> Optional[Piece]
            Gets the piece at the given position.

        make_move(move) -> GameState
            Makes the given move.
        step() -> GameState
            Advances a step of the game.
        run() -> GameState
            Runs the game until it's over.

        restart() -> GameState
            Returns the starting state with the same players.
    """

    __game_state_counter = count()

    def __init__(
        self,
        key,
        *,
        white: AnyPlayer,
        black: AnyPlayer,
        current_color: Color,
        prev: Optional[Self],
        move: Optional[Move],
        piece_captured: Optional[Piece],
        en_passant_pawn: Optional[Piece],
        half_move_clock: int,
        num_moves: int,
        board: BoardDict,
        captured: Iterable[Piece],
    ):
        """Initializes a GameState. Should only be called internally."""
        if key is not _PRIVATE_game_state_constructor_key:
            raise RuntimeError(
                "You should not be calling the GameState constructor"
            )

        if white.color is not Color.WHITE:
            raise ValueError("`white` does not have color `Color.WHITE`")
        if black.color is not Color.BLACK:
            raise ValueError("`black` does not have color `Color.BLACK`")
        self._white = white
        self._black = black
        self._current_color = current_color

        # cache the resulting game state whenever someone tries to make
        # a move, in case a bot wants to look ahead
        self._moves: Dict[Move, GameState] = {}
        # cache the resulting game state whenever a pawn is promoted
        self._promotions: Dict[PromoteType, GameState] = {}

        if (prev is None) != (move is None):
            raise ValueError("`prev` and `move` must both be None or not None")

        if prev is None:
            self._first_state = self
        else:
            self._first_state = prev._first_state
        self._prev = prev

        self._half_move_clock = half_move_clock
        self._num_moves = num_moves

        self._captured = tuple(captured)

        # construct board
        self._board: BoardDict = {}
        self._kings = [None, None]
        seen_piece_ids = set()
        unmoved_rooks = ([], [])
        for piece in board.values():
            if piece.id in seen_piece_ids:
                raise ValueError(f"Multiple pieces with id: {piece.id}")
            seen_piece_ids.add(piece.id)
            if piece.pos in self._board:
                raise ValueError(f"Multiple pieces at position {piece.pos}")
            self._board[piece.pos] = piece
            if piece.type is PieceType.KING:
                self._kings[piece.color.value] = piece
            elif piece.type is PieceType.ROOK:
                if not piece.has_moved:
                    unmoved_rooks[piece.color.value].append(piece)

        # validate last move
        piece_moved = None
        self._promoting_pawn = None
        move = Move.of(move)
        if move is None:
            self._move = None
        else:
            if self.get_piece(*move.pos) is not None:
                raise ValueError(
                    "Invalid move from prev: start position still has a piece"
                )
            piece_moved = self.get_piece(*move.result_pos)
            if piece_moved is None:
                raise ValueError(
                    "Invalid move from prev: end position is empty"
                )
            if piece_moved.type is PieceType.PAWN:
                if piece_moved.can_promote:
                    self._promoting_pawn = piece_moved

            # change from `Move` to `PieceMoved`
            self._move = PieceMoved(*move, piece_moved, piece_captured)

        # validate kings
        if self._kings == [None, None]:
            raise ValueError("Missing both kings")
        for i, king in enumerate(self._kings):
            if king is None:
                raise ValueError(f"Missing {Color(i).name.lower()} king")
        # make sure kings are not next to each other
        # board number doesn't matter, since being next to each other on
        # opposite boards meant someone was able to go close on the same
        # board, which is still invalid
        _, r0, c0 = self._kings[0].pos
        _, r1, c1 = self._kings[1].pos
        if abs(r0 - r1) <= 1 and abs(c0 - c1) <= 1:
            raise ValueError("Kings are next to each other")

        self._id = next(self.__game_state_counter)

        # compute fen representation
        # forsyth-edwards notation (FEN): https://www.chess.com/terms/fen-chess
        fen = []

        def color_case(color, value):
            if color is Color.WHITE:
                return value.upper()
            else:  # color is Color.BLACK
                return value.lower()

        # piece placements
        placements = []
        for bn in range(2):
            for r in range(8):
                rank = []
                empty = 0
                for c in range(8):
                    piece = self.get_piece(bn, r, c)
                    if piece is None:
                        empty += 1
                        continue
                    if empty > 0:
                        rank.append(str(empty))
                        empty = 0
                    rank.append(piece.fen_name)
                if empty > 0:
                    rank.append(str(empty))
                placements.append("".join(rank))
        fen.append("/".join(placements))
        # active color
        fen.append(self._current_color.abbr().lower())
        # castling rights
        castling = []
        for color in (Color.WHITE, Color.BLACK):
            king = self._kings[color.value]
            if king.has_moved:
                continue
            kc = king.pos.c
            kingside = False
            queenside = False
            for rook in unmoved_rooks[color.value]:
                rc = rook.pos.c
                if rc < kc:  # left rook: queenside
                    queenside = True
                else:  # right rook: kingside
                    kingside = True
            if kingside:
                # kingside
                castling.append(color_case(color, "K"))
            if queenside:
                # queenside
                castling.append(color_case(color, "Q"))
        if len(castling) == 0:
            fen.append("-")
        else:
            fen.append("".join(castling))
        # en passant target
        if en_passant_pawn is None:
            fen.append("-")
        else:
            _, r, c = en_passant_pawn.pos
            r -= en_passant_pawn.dr
            fen.append(Position(r, c).code)
        # half move clock
        fen.append(str(self._half_move_clock))
        # full move number
        fen.append(str(self._num_moves))
        # combine
        self._fen = " ".join(fen)

        if len(self._board) == 2:
            # only two kings are left
            self._is_in_check = False
            self._end_game_state = EndGameState.ONLY_KINGS_DRAW
            return
        if self._half_move_clock >= 100:
            # 50 move rule
            self._is_in_check = False
            self._end_game_state = EndGameState.FIFTY_MOVE_DRAW
            return

        # calculate all the possible moves
        calculator = MovesCalculator(
            self._board, self._kings, unmoved_rooks, en_passant_pawn
        )

        self._is_in_check = calculator.is_in_check(self._current_color)

        # determine end game state for current player
        no_more_moves = calculator.num_moves(self._current_color) == 0
        if no_more_moves:
            if self._is_in_check:
                self._end_game_state = EndGameState.CHECKMATE
            else:
                self._end_game_state = EndGameState.STALEMATE
        else:
            self._end_game_state = None

    def debug(self):
        print("=" * 50)
        print(self.fen())
        print(self.board_to_str())
        print(self.moves_to_str())
        if self.is_game_over():
            print("end game state:", self._end_game_state)
            if self.is_in_checkmate():
                print("winner:", self.winner())
        else:
            print("in check:", self.is_in_check())

    @classmethod
    def new(cls, *, white: Type[AnyPlayer], black: Type[AnyPlayer]) -> Self:
        """Initializes the first GameState of a game."""
        initial_fen = (
            # initial board position
            "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR/8/8/8/8/8/8/8/8",
            # white goes first
            "w",
            # all pieces can castle
            "KQkq",
            # no en passant
            "-",
            # no moves yet
            "0",
            "0",
        )
        return cls.from_fen(" ".join(initial_fen), white=white, black=black)

    @classmethod
    def from_fen(
        cls, fen: str, *, white: Type[AnyPlayer], black: Type[AnyPlayer]
    ) -> Self:
        """Initializes a GameState from a FEN string, using the given
        players.
        """

        white_player = white(Color.WHITE)
        black_player = black(Color.BLACK)

        fen_parts = fen.split()  # split by spaces
        if len(fen_parts) != 6:
            raise ValueError(f"`fen` expects 6 fields; got {len(fen_parts)}")
        (
            placements_str,
            active_color,
            castling_rights,
            en_passant_target,
            half_move_clock_str,
            full_move_number_str,
        ) = fen_parts

        # get pieces
        placements = placements_str.split("/")
        if len(placements) != 16:
            raise ValueError(
                "`fen` piece placements expects 16 ranks; "
                f"got {len(placements)}"
            )
        PIECE_CLASSES: Dict[str, Type[Piece]] = {
            "K": pieces.King,
            "Q": pieces.Queen,
            "R": pieces.Rook,
            "N": pieces.Knight,
            "B": pieces.Bishop,
            "P": pieces.Pawn,
        }
        piece_id = count()
        board: BoardDict = {}
        kings = [None, None]
        rooks = ([], [])
        for i, rank in enumerate(placements):
            bn, r = divmod(i, 8)
            num_files = 0
            for c in rank:
                # allow multiple digits in a row, just because i'm not
                # gonna be super strict about this
                if c.isdigit():
                    num_files += int(c)
                    continue
                piece_cls = PIECE_CLASSES.get(c.upper(), None)
                if piece_cls is None:
                    raise ValueError(f"Rank {i}: invalid piece symbol: {c!r}")
                color = Color.WHITE if c.isupper() else Color.BLACK
                piece = piece_cls(next(piece_id), color, (bn, r, num_files))
                board[piece.pos] = piece
                num_files += 1

                if piece.type is PieceType.KING:
                    kings[color.value] = piece
                elif piece.type is PieceType.ROOK:
                    rooks[color.value].append(piece)
            if num_files < 8:
                raise ValueError(f"Rank {i} has less than 8 files: {rank}")
            elif num_files > 8:
                raise ValueError(f"Rank {i} has more than 8 files: {rank}")

        # get active color
        if active_color == "w":
            current_color = Color.WHITE
        elif active_color == "b":
            current_color = Color.BLACK
        else:
            raise ValueError(f"Invalid active color symbol: {active_color!r}")

        # castling rights
        can_castle = ([False, False], [False, False])
        if castling_rights == "-":
            # nobody can castle
            pass
        else:
            for c in castling_rights:
                color = Color.WHITE if c.isupper() else Color.BLACK
                if c.upper() == "K":
                    # kingside castle
                    index = 1
                elif c.upper() == "Q":
                    # queenside castle
                    index = 0
                else:
                    raise ValueError(f"Invalid castling symbol: {c!r}")
                can_castle[color.value][index] = True
        for color in Color:
            # pylint: disable=protected-access
            ability = can_castle[color.value]
            if ability == [True, True]:
                # can castle on both sides, so leave the pieces
                # untouched
                continue
            king = kings[color.value]
            if king is None:
                # ignore; the error will come up later
                continue
            if ability == [False, False]:
                # can't castle on either side, so just say the king
                # moved
                king._set_moved()
                continue
            kc = king.pos.c
            for rook in rooks[color.value]:
                if rook.has_moved:
                    # not in an initial position, so just say it moved
                    rook._set_moved()
                    continue
                rc = rook.pos.c
                if rc < kc:  # left rook: queenside
                    index = 0
                else:  # right rook:
                    index = 1
                if not ability[index]:
                    # can't castle this way
                    rook._set_moved()

        # get en passant pawn
        if en_passant_target == "-":
            en_passant_pawn = None
        else:
            try:
                en_passant_pos = Position.from_code(en_passant_target)
            except ValueError as e:
                raise ValueError(
                    f"Invalid en passant target: {en_passant_target!r}"
                ) from e
            r, c = en_passant_pos
            # check for a pawn that is right in front of this position
            if r == 2:
                pawn = board.get((1, r + 1, c), None)
            elif r == 5:
                pawn = board.get((1, r - 1, c), None)
            else:
                raise ValueError(
                    "Invalid en passant target: rank must be 3 (for black) or "
                    "6 (for white)"
                )
            if pawn is None:
                raise ValueError(
                    f"No pawn found for en passant target: {en_passant_target}"
                )
            if pawn.type is not PieceType.PAWN:
                raise ValueError(
                    f"Piece for en passant target {en_passant_pos} is not a "
                    f"pawn: {pawn.type.title()} at {pawn.pos}"
                )
            # check that the pawn could have made this move in the last
            # turn
            if any(
                pos in board
                for pos in (
                    # something was blocking on the original board
                    (0, r, c),
                    # something else is in the original position
                    (0, r - pawn.dr, c),
                    (1, r - pawn.dr, c),
                )
            ):
                raise ValueError(
                    f"En passant pawn at {pawn.pos} could not have advanced "
                    "two squares in the last move"
                )
            en_passant_pawn = pawn

        if not half_move_clock_str.isdigit():
            raise ValueError(
                f"Halfmove clock is not an integer: {half_move_clock_str!r}"
            )
        half_move_clock = int(half_move_clock_str)
        if not full_move_number_str.isdigit():
            raise ValueError(
                f"Fullmove number is not an integer: {full_move_number_str!r}"
            )
        num_moves = int(full_move_number_str)

        return cls(
            _PRIVATE_game_state_constructor_key,
            white=white_player,
            black=black_player,
            current_color=current_color,
            # we don't know anything about the previous moves
            prev=None,
            move=None,
            piece_captured=None,
            en_passant_pawn=en_passant_pawn,
            half_move_clock=half_move_clock,
            num_moves=num_moves,
            board=board,
            # we don't know anything about the captured pieces
            captured=[],
        )

    def __str__(self) -> str:
        return self._fen

    @property
    def id(self) -> int:
        """The game state id."""
        return self._id

    @property
    def white(self) -> AnyPlayer:
        """The white player."""
        return self._white

    @property
    def black(self) -> AnyPlayer:
        """The black player."""
        return self._black

    @property
    def prev(self) -> Optional[Self]:
        """The previous GameState, or None if this is the first state."""
        return self._prev

    @property
    def move(self) -> Optional[PieceMoved]:
        """The move from the previous GameState, or None if this is the
        first state.
        """
        return self._move

    @property
    def current_color(self) -> Color:
        """The color of the current player."""
        return self._current_color

    @property
    def current_player(self) -> AnyPlayer:
        """The current player."""
        if self._current_color is Color.WHITE:
            return self._white
        else:
            return self._black

    @property
    def half_move_clock(self) -> int:
        """The number of half-moves since the last capture or the last
        pawn moved.
        """
        return self._half_move_clock

    @property
    def num_moves(self):
        """The number of full moves played in the entire game."""
        return self._num_moves

    def fen(self) -> str:
        """Returns the FEN representation of the game.

        Since Alice Chess has two boards, the piece placement portion is
        doubled in length, where the first 8 ranks refer to Board A (on
        the left) and the last 8 ranks refer to Board B (on the right).
        """
        return self._fen

    def board_to_str(self, empty: str = ".") -> str:
        """Returns a string representation of the board.

        Args:
            empty (str): A placeholder to use for empty positions.
                Must have length 1.

        Raises:
            ValueError: If `empty` does not have length 1.

        Returns:
            str: The string representation.
        """
        if len(empty) != 1:
            raise ValueError("`empty` must have length 1")
        board_str = []
        for r in range(8):
            line = []
            for bn in range(2):
                for c in range(8):
                    pos = (bn, r, c)
                    if pos not in self._board:
                        line.append(empty)
                    else:
                        piece = self._board[pos]
                        line.append(piece.fen_name)
                line.append(" ")
            board_str.append(" ".join(line))
        return "\n".join(board_str)

    def moves_to_str(
        self, columns: int = 4, captured_empty: bool = True
    ) -> str:
        """Returns a string representation of the moves for each piece.

        If `captured_empty` is True, there will be empty spaces for
        piece ids that don't have associated pieces currently on the
        board (assumed to be captured).

        Args:
            columns (int): The number of columns to split the pieces
                into. Must be positive.
            captured_empty (bool): Whether to leave empty spaces for
                captured pieces.

        Raises:
            ValueError: If `columns` is not positive.

        Returns:
            str: The string representation.
        """
        COL_PADDING = 4
        if columns <= 0:
            raise ValueError("`columns` must be positive")
        if captured_empty:
            pieces_by_id = {
                piece.id: piece for piece in self.yield_all_pieces()
            }
            all_pieces = [
                (i, pieces_by_id.get(i, None))
                for i in range(max(pieces_by_id.keys()) + 1)
            ]
        else:
            all_pieces = sorted(
                (piece.id, piece) for piece in self.yield_all_pieces()
            )
        num_rows = math.ceil(len(all_pieces) / columns)
        col_widths = [0] * columns
        col_has_piece = [False] * columns
        cols = [[] for _ in range(columns)]
        col_i = 0
        for piece_id, piece in all_pieces:
            if piece is None:
                line = f"({piece_id})"
            else:
                args = (
                    f"({piece_id})",
                    piece.name,
                    piece.pos,
                    ":",
                    " ".join(str(move.target) for move in piece.yield_moves()),
                )
                line = " ".join(map(str, args))
                col_has_piece[col_i] = True
            col = cols[col_i]
            col.append(line)
            col_widths[col_i] = max(len(line), col_widths[col_i])
            if len(col) >= num_rows:
                # start filling up the next column
                col_i += 1
        return "\n".join(
            (" " * COL_PADDING).join(
                value.ljust(width)
                for value, include_col, width in zip(
                    row, col_has_piece, col_widths
                )
                if include_col
            )
            # use zip longest because the last column may be shorter
            for row in zip_longest(*cols, fillvalue="")
        )

    def is_game_over(self) -> bool:
        """Returns whether the game is over (checkmate, stalemate, or
        draw).
        """
        return self._end_game_state is not None

    def is_in_checkmate(self) -> bool:
        """Returns whether the current player is in checkmate."""
        return self._end_game_state is EndGameState.CHECKMATE

    def winner(self) -> Optional[Color]:
        """Returns the winner of the game (for checkmate)."""
        if not self.is_in_checkmate():
            return None
        # the winner is the other color, since the current color is in
        # checkmate
        return self._current_color.other()

    def is_in_stalemate(self) -> bool:
        """Returns whether the current player is in stalemate."""
        return self._end_game_state is EndGameState.STALEMATE

    def is_kings_draw(self) -> bool:
        """Returns whether the game is a draw by only having kings."""
        return self._end_game_state is EndGameState.ONLY_KINGS_DRAW

    def is_draw(self) -> bool:
        """Returns whether the game is a draw (by the fifty move rule)."""
        return self._end_game_state is EndGameState.FIFTY_MOVE_DRAW

    def is_in_check(self) -> bool:
        """Returns whether the current player is in check."""
        return self._is_in_check

    def needs_promotion(self) -> bool:
        """Returns whether the state is waiting for a promotion for the
        last moved pawn.
        """
        return self._promoting_pawn is not None

    def promote(self, promote_type: PromoteType) -> Self:
        """Promotes the promoting pawn.

        A pawn is able to be promoted if it was last moved into the last
        row for its direction of movement.

        Args:
            promote_type (PromoteType): The type to promote to.

        Raises:
            RuntimeError: If there is no pawn waiting for a promotion.

        Returns:
            GameState: The new game state with the promoted pawn.
        """
        if not self.needs_promotion():
            raise RuntimeError("No pawn waiting for a promotion")

        if promote_type in self._promotions:
            return self._promotions[promote_type]

        pawn = self._promoting_pawn

        # create copy of board
        board = {pos: piece.copy() for pos, piece in self._board.items()}

        # replace this pawn with the specified piece type
        promoted_piece = make_promoted(pawn, promote_type)
        board[pawn.pos] = promoted_piece

        # create new game state
        new_state = self.__class__(
            _PRIVATE_game_state_constructor_key,
            white=self._white,
            black=self._black,
            current_color=self._current_color,
            # direct back to this state's previous state
            prev=self._prev,
            # use the same move from previous
            move=(self._move.pos, self._move.target),
            piece_captured=self._move.piece_captured,
            # this pawn was promoting, so there is no en passant pawn
            en_passant_pawn=None,
            half_move_clock=self._half_move_clock,
            num_moves=self._num_moves,
            board=board,
            captured=self._captured,
        )
        # cache the promotion state
        self._promotions[promote_type] = new_state
        return new_state

    def yield_all_pieces(self) -> Iterator[Piece]:
        """Yields all the pieces."""
        return iter(self._board.values())

    def yield_player_pieces(self) -> Iterator[Piece]:
        """Yields all the pieces for the current player."""
        for piece in self.yield_all_pieces():
            if piece.color is self._current_color:
                yield piece

    def yield_player_moves(self) -> Iterator[PieceMove]:
        """Yields all the possible moves for the current player."""
        for piece in self.yield_player_pieces():
            for piece_move in piece.yield_moves():
                yield piece_move

    def num_captured(self) -> int:
        """Returns the number of captured pieces."""
        return len(self._captured)

    def yield_captured(self) -> Iterator[Piece]:
        """Yields the captured pieces."""
        return iter(self._captured)

    def get_piece(self, bn: int, r: int, c: int) -> Optional[Piece]:
        """Gets the piece at the given position.

        Args:
            bn (int): The board number.
            r (int): The row.
            c (int): The column.

        Raises:
            ValueError:
                If `bn` is not 0 or 1.
                If `r` and `c` are not bounded within [0, 7].

        Returns:
            Optional[Piece]: The piece, or None.
        """
        return self._board.get((bn, r, c), None)

    def make_move(self, move: Move) -> Self:
        """Makes the given move.

        If a promotion is needed after this move, the pawn will simply
        be moved without being promoted. It is up to the caller to call
        the `GameState.promote()` method on the returned state.

        Args:
            move (Move): The move to make.

        Raises:
            RuntimeError:
                If the game is over.
                If the state is waiting for a promotion.
            ValueError: If the given move is invalid.

        Returns:
            GameState: The next game state.
        """
        move = Move.of(move)
        (bn, r, c), (tr, tc) = move

        if self.is_game_over():
            raise RuntimeError("Game is over")
        if self.needs_promotion():
            raise RuntimeError("Game needs a promotion")

        if move in self._moves:
            return self._moves[move]

        piece = self.get_piece(*move.pos)
        if piece is None:
            raise ValueError(f"No piece at position {move.pos}")
        if piece.color is not self._current_color:
            raise ValueError("Cannot move opponent's piece")
        if not piece.can_make_move(*move.capture_pos):
            raise ValueError(
                f"Piece at {move.pos} cannot move to {move.target}"
            )

        # check just in case
        other_board = self.get_piece(*move.result_pos)
        if other_board is not None:
            raise ValueError("Move squashes a piece on the other board")

        # make new copy of board
        board = {pos: piece.copy() for pos, piece in self._board.items()}
        captured = list(self._captured)

        half_move_clock = self._half_move_clock + 1

        # make move
        moved_piece = board.pop(piece.pos).move_to(move.result_pos)
        board[move.result_pos] = moved_piece

        piece_captured = board.pop(move.capture_pos, None)
        # check for en passant
        if piece.type is PieceType.PAWN:
            # reset clock: pawn moved
            half_move_clock = 0
            if c == tc:
                # going straight
                pass
            elif piece_captured is not None:
                # regular capture (diagonally)
                pass
            else:
                # change capture to en passant (same row)
                piece_captured = board.pop((bn, r, tc), None)
                if piece_captured is None:
                    raise ValueError(
                        "Pawn performed en passant, but no piece to capture"
                    )
                if piece_captured.type is not PieceType.PAWN:
                    raise ValueError("Cannot perform en passant on non-pawn")
        # capture piece
        if piece_captured is not None:
            # reset clock: piece was captured
            half_move_clock = 0
            piece_captured = piece_captured.capture()
            captured.append(piece_captured)

        # check for castle
        if piece.type is PieceType.KING:
            if not piece.has_moved and r == tr and abs(c - tc) == 2:
                # king is castling
                if tc < c:
                    if not piece.can_castle_left:
                        raise ValueError("King cannot castle")
                    # castling to the left (queenside)
                    rook_old_c = 0
                    rook_new_c = piece.left_rook_col
                else:  # tc > c
                    if not piece.can_castle_right:
                        raise ValueError("King cannot castle")
                    # castling to the right (kingside)
                    rook_old_c = 7
                    rook_new_c = piece.right_rook_col
                # move rook to new board between old and new positions
                # of king
                rook_old_pos = (bn, tr, rook_old_c)
                rook_new_pos = (1 - bn, tr, rook_new_c)
                board[rook_new_pos] = board.pop(rook_old_pos).move_to(
                    rook_new_pos
                )

        # check for en passant
        en_passant_pawn = None
        if piece.type is PieceType.PAWN and not piece.has_moved:
            if bn == 0 and abs(r - tr) == 2 and c == tc:
                # pawn advanced two steps
                en_passant_pawn = moved_piece

        num_moves = self._num_moves
        if self._current_color is Color.BLACK:
            # increments when black plays
            num_moves += 1

        # create next state
        next_state = self.__class__(
            _PRIVATE_game_state_constructor_key,
            white=self._white,
            black=self._black,
            current_color=self._current_color.other(),
            prev=self,
            move=move,
            piece_captured=piece_captured,
            en_passant_pawn=en_passant_pawn,
            half_move_clock=half_move_clock,
            num_moves=num_moves,
            board=board,
            captured=captured,
        )
        self._moves[move] = next_state
        return next_state

    def step(self) -> Self:
        """Advances a step of the game.

        Prompts the current player for a move, and also prompts them for
        promotion if necessary. Returns the next game state.

        Raises:
            RuntimeError: If the current player is a human (call
                `make_move()` instead).
            ValueError: If the move given by the player is invalid.

        Returns:
            GameState: The next game state.
        """

        player = self.current_player
        if player.is_human:
            raise RuntimeError(
                "Current player is a human (call `make_move()` instead)"
            )

        # ask player to make move
        move = player.make_move(self)
        next_state = self.make_move(move)

        if next_state.needs_promotion():
            # ask player to make promotion
            promote_type = player.promote(self)
            next_state = next_state.promote(promote_type)

        return next_state

    def run(self) -> Self:
        """Runs the game until it's over.

        Returns the last game state.
        """
        state = self
        while not state.is_game_over():
            state = state.step()
        return state

    def restart(self) -> Self:
        """Returns the starting state with the same players."""
        return self._first_state



"""
Utility constants and functions.
"""

# =============================================================================

import enum as _enum

# =============================================================================

__all__ = (
    "Color",
    "PieceType",
    "PromoteType",
    "EndGameState",
    "check_brc",
    "check_brc_bool",
)

# =============================================================================


class _EnumWIthTitle(_enum.Enum):
    """An enum class that adds a `title()` method."""

    def title(self) -> str:
        """Returns the enum name as a title string."""
        return self.name.title()


class Color(_EnumWIthTitle):
    """The colors for each piece."""

    WHITE = 0
    BLACK = 1

    def abbr(self) -> str:
        """Returns the first letter of the color."""
        return self.name[0]

    def other(self) -> "Color":
        """Returns the other color."""
        return Color(1 - self.value)


class PieceType(_EnumWIthTitle):
    """The piece types."""

    KING = "K"
    QUEEN = "Q"
    ROOK = "R"
    KNIGHT = "N"
    BISHOP = "B"
    PAWN = "P"


class PromoteType(_enum.Enum):
    """The piece types that can be promoted to."""

    QUEEN = PieceType.QUEEN
    ROOK = PieceType.ROOK
    KNIGHT = PieceType.KNIGHT
    BISHOP = PieceType.BISHOP

    @classmethod
    def by_index(cls, i: int) -> "PromoteType":
        """Returns the PromoteType at index i."""
        for j, p_type in enumerate(cls):
            if j == i:
                return p_type
        raise IndexError(f"invalid index {i} for {cls.__name__}")


class EndGameState(_enum.Enum):
    """The possible states for the end of the game."""

    CHECKMATE = _enum.auto()
    STALEMATE = _enum.auto()
    ONLY_KINGS_DRAW = _enum.auto()
    FIFTY_MOVE_DRAW = _enum.auto()


# =============================================================================


def check_brc(bn=None, r=None, c=None, tr=None, tc=None):
    """Checks that all the given args are within the proper bounds.

    Raises `ValueError` if any conditions are not satisfied.
    """
    if bn is not None and bn not in (0, 1):
        raise ValueError("`bn` must be 0 or 1")
    for name, val in (("r", r), ("c", c), ("tr", tr), ("tc", tc)):
        if val is None:
            continue
        if not 0 <= val < 8:
            raise ValueError(f"`{name}` must be bounded within [0, 7]")


def check_brc_bool(bn=None, r=None, c=None, tr=None, tc=None) -> bool:
    """Returns True if all the given args are within the proper bounds."""
    try:
        check_brc(bn, r, c, tr, tc)
        return True
    except ValueError:
        return False
